@@ -1,6 +1,6 @@
 /*
  *   stunnel       TLS offloading and load-balancing proxy
- *   Copyright (C) 1998-2016 Michal Trojnara <Michal.Trojnara@stunnel.org>
+ *   Copyright (C) 1998-2019 Michal Trojnara <Michal.Trojnara@stunnel.org>
  *
  *   This program is free software; you can redistribute it and/or modify it
  *   under the terms of the GNU General Public License as published by the
@@ -108,7 +108,11 @@ NOEXPORT int main_unix(int argc, char* argv[]) {
             signal(SIGINT, signal_handler); /* fatal */
 #endif
         daemon_loop();
+#if !defined(__vms) && !defined(USE_OS2)
+        delete_pid();
+#endif /* standard Unix */
     } else { /* inetd mode */
+        CLI *c;
 #if !defined(__vms) && !defined(USE_OS2)
         close(fd);
 #endif /* standard Unix */
@@ -118,7 +122,10 @@ NOEXPORT int main_unix(int argc, char* argv[]) {
 #endif
         set_nonblock(0, 1); /* stdin */
         set_nonblock(1, 1); /* stdout */
-        client_main(alloc_client_session(&service_options, 0, 1));
+        c=alloc_client_session(&service_options, 0, 1);
+        tls_alloc(c, ui_tls, NULL);
+        service_up_ref(&service_options);
+        client_main(c);
     }
     return 0;
 }
@@ -128,7 +135,7 @@ NOEXPORT void signal_handler(int sig) {
     int saved_errno;
 
     saved_errno=errno;
-    signal_post(sig);
+    signal_post((uint8_t)sig);
     signal(sig, signal_handler);
     errno=saved_errno;
 }
@@ -176,22 +183,18 @@ NOEXPORT int create_pid(void) {
         s_log(LOG_DEBUG, "No pid file being created");
         return 0;
     }
-    if(global_options.pidfile[0]!='/') {
-        /* to prevent creating pid file relative to '/' after daemonize() */
-        s_log(LOG_ERR, "Pid file (%s) must be full path name", global_options.pidfile);
-        return 1;
-    }
-    global_options.dpid=(unsigned long)getpid();
 
-    /* silently remove old pid file */
+    /* silently remove the old pid file */
     unlink(global_options.pidfile);
+
+    /* create a new pid file */
     pf=open(global_options.pidfile, O_WRONLY|O_CREAT|O_TRUNC|O_EXCL, 0644);
     if(pf==-1) {
         s_log(LOG_ERR, "Cannot create pid file %s", global_options.pidfile);
         ioerror("create");
         return 1;
     }
-    pid=str_printf("%lu\n", global_options.dpid);
+    pid=str_printf("%lu\n", (unsigned long)getpid());
     if(write(pf, pid, strlen(pid))<(int)strlen(pid)) {
         s_log(LOG_ERR, "Cannot write pid file %s", global_options.pidfile);
         ioerror("write");
@@ -200,16 +203,18 @@ NOEXPORT int create_pid(void) {
     str_free(pid);
     close(pf);
     s_log(LOG_DEBUG, "Created pid file %s", global_options.pidfile);
-    atexit(delete_pid);
     return 0;
 }
 
 NOEXPORT void delete_pid(void) {
-    if((unsigned long)getpid()!=global_options.dpid)
-        return; /* current process is not main daemon process */
-    s_log(LOG_DEBUG, "removing pid file %s", global_options.pidfile);
-    if(unlink(global_options.pidfile)<0)
-        ioerror(global_options.pidfile); /* not critical */
+    if(global_options.pidfile) {
+        if(unlink(global_options.pidfile)<0)
+            ioerror(global_options.pidfile); /* not critical */
+        else
+            s_log(LOG_DEBUG, "Removed pid file %s", global_options.pidfile);
+    } else {
+        s_log(LOG_DEBUG, "No pid file to remove");
+    }
 }
 
 #endif /* standard Unix */
@@ -252,9 +257,8 @@ void ui_new_log(const char *line) {
 
 /**************************************** ctx callbacks */
 
-int passwd_cb(char *buf, int size, int rwflag, void *userdata) {
-    (void)userdata; /* squash the unused parameter warning */
-    return PEM_def_callback(buf, size, rwflag, NULL);
+int ui_passwd_cb(char *buf, int size, int rwflag, void *userdata) {
+    return PEM_def_callback(buf, size, rwflag, userdata);
 }
 
 #ifndef OPENSSL_NO_ENGINE
